'''Base58 encoding

Implementations of Base58 and Base58Check endcodings that are compatible
with the bitcoin network.
'''

from hashlib import sha256

# 58 character alphabet used
alphabet = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'

iseq = lambda s: s
bseq = bytes
buffer = lambda s: s.buffer

def b58encode(v):
    '''Encode a string using Base58'''

    if not isinstance(v, bytes):
        raise TypeError("a bytes-like object is required, not '%s'" %
                        type(v).__name__)

    origlen = len(v)
    v = v.lstrip(b'\0')
    newlen = len(v)

    p, acc = 1, 0
    for c in iseq(v[::-1]):
        acc += p * c
        p = p << 8

    result = ''
    while acc > 0:
        acc, mod = divmod(acc, 58)
        result += alphabet[mod]

    return (result + alphabet[0] * (origlen - newlen))[::-1]


def b58decode(v):
    '''Decode a Base58 encoded string'''

    if not isinstance(v, str):
        v = v.decode('ascii')

    origlen = len(v)
    v = v.lstrip(alphabet[0])
    newlen = len(v)

    p, acc = 1, 0
    for c in v[::-1]:
        acc += p * alphabet.index(c)
        p *= 58

    result = []
    while acc > 0:
        acc, mod = divmod(acc, 256)
        result.append(mod)

    return (bseq(result) + b'\0' * (origlen - newlen))[::-1]


def b58encode_check(v):
    '''Encode a string using Base58 with a 4 character checksum'''

    digest = sha256(sha256(v).digest()).digest()
    return b58encode(v + digest[:4])


def b58decode_check(v):
    '''Decode and verify the checksum of a Base58 encoded string'''

    result = b58decode(v)
    result, check = result[:-4], result[-4:]
    digest = sha256(sha256(result).digest()).digest()

    if check != digest[:4]:
        raise ValueError("Invalid checksum")

    return result





'''Communicate with local or remote peercoin-daemon via JSON-RPC'''

from operator import itemgetter

try:
    from peercoin_rpc import Client
except:
    raise EnvironmentError("peercoin_rpc library is required for this to work,\
                            use pip to install it.")

class RpcNode(Client):
    '''JSON-RPC connection to local Peercoin node'''

    def select_inputs(self, total_amount, address=None):
        '''finds apropriate utxo's to include in rawtx, while being careful
        to never spend old transactions with a lot of coin age.
        Argument is intiger, returns list of apropriate UTXO's'''

        utxo = []
        utxo_sum = float(-0.01) ## starts from negative due to minimal fee
        for tx in sorted(self.listunspent(address=address), key=itemgetter('confirmations')):

            utxo.append({
                "txid": tx["txid"],
                "vout": tx["vout"],
                "scriptSig": tx["scriptPubKey"],
                "amount": tx["amount"],
                "address": tx["address"]
            })

            utxo_sum += float(tx["amount"])
            if utxo_sum >= total_amount:
                return {'utxos':utxo, 'total':utxo_sum}

        if utxo_sum < total_amount:
            raise ValueError("Not enough funds.")

    @property
    def is_testnet(self):
        '''check if node is configured to use testnet or mainnet'''

        if self.getinfo()["testnet"] is True:
            return True
        else:
            return False

    @property
    def network(self):
        '''return which network is the node operating on.'''

        if self.is_testnet:
            return "tppc"
        else:
            return "ppc"

    def listunspent(self, minconf=1, maxconf=999999, address=None):
        '''list UTXOs
        modified version to allow filtering by address.
        '''
        if address:
            return [u for u in self.req("listunspent", [minconf, maxconf]) if u["address"] == address]
        else:
            return self.req("listunspent", [minconf, maxconf])





'''various contants'''

### P2TH tags for Peercoin mainnet/testnet ### ### ### 

## mainnet
mainnet_PAPROD = "U624wXL6iT7XZ9qeHsrtPGEiU78V1YxDfwq75Mymd61Ch56w47KE"
mainnet_PAPROD_addr = "PAprodbYvZqf4vjhef49aThB9rSZRxXsM6"
mainnet_PATEST = "UAbxMGQQKmfZCwKXAhUQg3MZNXcnSqG5Z6wAJMLNVUAcyJ5yYxLP"
mainnet_PATEST_addr = "PAtesth4QreCwMzXJjYHBcCVKbC4wjbYKP"

## testnet
testnet_PAPROD = "cTJVuFKuupqVjaQCFLtsJfG8NyEyHZ3vjCdistzitsD2ZapvwYZH"
testnet_PAPROD_addr = "miHhMLaMWubq4Wx6SdTEqZcUHEGp8RKMZt"
testnet_PATEST = "cQToBYwzrB3yHr8h7PchBobM3zbrdGKj2LtXqg7NQLuxsHeKJtRL"
testnet_PATEST_addr = "mvfR2sSxAfmDaGgPcmdsTwPqzS6R9nM5Bo"

### ### ### ### ### ### ### ### ### ### ### ### ### ###

### P2TH fee

mainnet_p2th_fee = float(0.01)
testnet_p2th_fee = float(0.01)





'''Mock provider, with usable but fake data, used for testing'''

def select_inputs(amount):
    '''returns list of UTXO's'''

    # inputs expected as [{'txid':txhash,'vout':index,'scriptSig':txinScript},..]

    return [{
        'txid': 'ff6739fefb883a6db24efa20d6440f65082cc0557d5f7f21d1a31afaa267ee67',
        'vout': 1,
        'scriptSig': '2102f7552f53d240ba22461351808041d267f941a2fd94e948e5817822459725abdeac'
    },
    {
        'txid': 'fb23ec93649080a848ba746d90c744db2e474797cbb9d3ad2c6aba08d8eff9b7',
        'vout': 2,
        'scriptSig': '210329706d16792f1b7661bb7f35660eb2a6858110dd5259ac84a12aa1838efea86dac'
        }
    ]


import requests

class Mintr:

    '''API wrapper for mintr.org blockexplorer,
    it only implements queries relevant to peerassets.
    This wrapper does some tweaks to output to match original RPC response.'''

    @classmethod
    def __init__(cls, network="peercoin"):

        cls.net = network
        cls.api = "https://{0}.mintr.org/api/".format(cls.net)

    @property
    def is_testnet(self):
        '''testnet or not?'''
        return False

    @property
    def network(self):
        '''which network is this running on?'''

        if self.network == "peercoin":
            return "ppc"

    @classmethod
    def get(cls, query):

        requests.packages.urllib3.disable_warnings()
        return requests.get(cls.api + query, verify=False).json()

    @classmethod
    def getinfo(cls):
        '''mock response, to allow compatibility with local rpc node'''

        return {"testnet": False}

    @classmethod
    def getrawtransaction(cls, txid, verbose=1):
        '''this mimics the behaviour of local node `getrawtransaction` query with argument 1'''

        def wrapper(raw):
            '''make Mintr API response just like RPC response'''

            raw["blocktime"] = raw["time"]
            raw.pop("time")

            for v in raw["vout"]:
                v["scriptPubKey"] = {"asm": v["asm"], "hex": v["hex"],
                                     "type": v["type"], "reqSigs": v["reqsigs"],
                                     "address": [v["address"]]
                                    }
                for k in ("address", "asm", "hex", "reqsigs", "type"):
                    v.pop(k)

            return raw

        if verbose == 0:
            return cls.get("tx/hash/" + txid)
        else:
            resp = cls.get("tx/hash/" + txid + "/full")
            if not resp == {'error': 'Unknown API call'}:
                return wrapper(resp)

    @classmethod
    def listtransactions(cls, addr):
        '''get information about <address>'''

        response = cls.get("address/balance/" + addr + "/full")
        assert response != {'error': 'Could not decode hash'}, {"error": "Can not find the address."}

        txid = []
        for i in response["transactions"]:
            t = {"confirmations": i["confirmations"],
                 "time": i["time"],
                 "txid": i["tx_hash"],
                 "address": response["address"]
                }
            if i["sent"] == "":
                t["amount"] = i["received"]
                t["category"] = "send"
            else:
                t["amount"] = i["sent"]
                t["category"] = "receive"

            txid.append(t)

        return txid




from collections import namedtuple

Network = namedtuple('Network', [
    'network_name',
    'network_shortname',
    'pubkeyhash',
    'wif_prefix',
    'scripthash',
    'magicbytes',
    'msgPrefix',
    'min_tx_fee',
    'tx_timestamp',
    'denomination',
])


'''
Network name should be capitalized, for testnet append "-testnet".
For abbreviation prefix testnet of the network with "t".
'''

networks = (
    # Peercoin mainnet
    Network("Peercoin", "ppc", b'37', b'b7', b'75', b'e6e8e9e5',
            b'\x17PPCoin Signed Message:\n', float(0.01), True, 1000000),
    # Peercoin testnet
    Network("Peercoin-testnet", "tppc", b'6f', b'ef', b'c4', b'cbf2c0ef',
            b'\x17PPCoin Signed Message:\n', float(0.01), True, 1000000),
    # Bitcoin mainnet
    Network("Bitcoin", "btc", b'00', b'80', b'05', b'd9b4bef9',
            b'\x18Bitcoin Signed Message:\n', 0, False, 100000000),
    # Bitcoin testnet
    Network("Bitcoin-testnet", "tbtc", b'6f', b'ef', b'c4', b'dab5bffa',
            b'\x18Bitcoin Signed Message:\n', 0, False, 100000000)
)

def query(query):
    '''find matching parameter among the networks'''

    for network in networks:
        for field in network:
            if field == query:
                return network



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: 0001-peerassets-transaction-specification.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='0001-peerassets-transaction-specification.proto',
  package='',
  syntax='proto3',
  serialized_pb=_b('\n/0001-peerassets-transaction-specification.proto\"\xb7\x01\n\tDeckSpawn\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x1a\n\x12number_of_decimals\x18\x03 \x01(\r\x12\x12\n\nissue_mode\x18\x04 \x01(\r\x12\x1b\n\x13\x61sset_specific_data\x18\x05 \x01(\x0c\">\n\x04MODE\x12\x08\n\x04NONE\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x08\n\x04ONCE\x10\x02\x12\t\n\x05MULTI\x10\x04\x12\x0b\n\x07SINGLET\x10\x08\"i\n\x0c\x43\x61rdTransfer\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x0f\n\x07\x61mounts\x18\x02 \x03(\x04\x12\x1a\n\x12number_of_decimals\x18\x03 \x01(\r\x12\x1b\n\x13\x61sset_specific_data\x18\x04 \x01(\x0c\x62\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)



_DECKSPAWN_MODE = _descriptor.EnumDescriptor(
  name='MODE',
  full_name='DeckSpawn.MODE',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NONE', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUSTOM', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ONCE', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MULTI', index=3, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SINGLET', index=4, number=8,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=173,
  serialized_end=235,
)
_sym_db.RegisterEnumDescriptor(_DECKSPAWN_MODE)


_DECKSPAWN = _descriptor.Descriptor(
  name='DeckSpawn',
  full_name='DeckSpawn',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='DeckSpawn.version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='DeckSpawn.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='number_of_decimals', full_name='DeckSpawn.number_of_decimals', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='issue_mode', full_name='DeckSpawn.issue_mode', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='asset_specific_data', full_name='DeckSpawn.asset_specific_data', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _DECKSPAWN_MODE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=52,
  serialized_end=235,
)


_CARDTRANSFER = _descriptor.Descriptor(
  name='CardTransfer',
  full_name='CardTransfer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='CardTransfer.version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='amounts', full_name='CardTransfer.amounts', index=1,
      number=2, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='number_of_decimals', full_name='CardTransfer.number_of_decimals', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='asset_specific_data', full_name='CardTransfer.asset_specific_data', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=237,
  serialized_end=342,
)

_DECKSPAWN_MODE.containing_type = _DECKSPAWN
DESCRIPTOR.message_types_by_name['DeckSpawn'] = _DECKSPAWN
DESCRIPTOR.message_types_by_name['CardTransfer'] = _CARDTRANSFER

DeckSpawn = _reflection.GeneratedProtocolMessageType('DeckSpawn', (_message.Message,), dict(
  DESCRIPTOR = _DECKSPAWN,
  __module__ = '0001_peerassets_transaction_specification_pb2'
  # @@protoc_insertion_point(class_scope:DeckSpawn)
  ))
_sym_db.RegisterMessage(DeckSpawn)

CardTransfer = _reflection.GeneratedProtocolMessageType('CardTransfer', (_message.Message,), dict(
  DESCRIPTOR = _CARDTRANSFER,
  __module__ = '0001_peerassets_transaction_specification_pb2'
  # @@protoc_insertion_point(class_scope:CardTransfer)
  ))
_sym_db.RegisterMessage(CardTransfer)


# @@protoc_insertion_point(module_scope)




'''transaction assembly/dissasembly'''

from pypeerassets.base58 import b58decode
from binascii import hexlify, unhexlify
from .networks import query, networks
from time import time
import struct

OP_RETURN = b'\x6a'
OP_PUSHDATA1 = b'\x4c'
OP_DUP = b'\x76'
OP_HASH160 = b'\xa9'
OP_EQUALVERIFY = b'\x88'
OP_CHECKSIG = b'\xac'
OP_1 = b'\x51'
OP_2 = b'\x52'
OP_3 = b'\x53'
OP_CHECKMULTISIG = b'\xae'
OP_EQUAL = b'\x87'

class Tx_buffer:
    '''helper class for unpacking binary data'''

    def __init__(self, data, ptr=0):
        self.data = data
        self.len = len(data)
        self.ptr = ptr

    def shift(self, chars):
        prefix = self.data[self.ptr:self.ptr+chars]
        self.ptr += chars

        return prefix

    def shift_unpack(self, chars, format):
        unpack = struct.unpack(format, self.shift(chars))

        return unpack[0]

    def shift_varint(self):
        value = self.shift_unpack(1, 'B')

        if value == 0xFF:
            value = self.shift_uint64()
        elif value == 0xFE:
            value = self.shift_unpack(4, '<L')
        elif value == 0xFD:
            value = self.shift_unpack(2, '<H')

        return value

    def shift_uint64(self):
        return self.shift_unpack(4, '<L') + 4294967296 * self.shift_unpack(4, '<L')

    def used(self):
        return min(self.ptr, self.len)

    def remaining(self):
        return max(self.len - self.ptr, 0)


def get_hash160(address):
    '''return ripemd160 hash of the pubkey form the address'''

    return b58decode(address)[1:-4]

def op_push(n):

    if n < 0x4c:
        return (n).to_bytes(1, byteorder='little')              # Push n bytes.
    elif n <= 0xff:
        return b'\x4c' + (n).to_bytes(1, byteorder='little')    # OP_PUSHDATA1
    elif n <= 0xffff:
        return b'\x4d' + (n).to_bytes(2, byteorder='little')    # OP_PUSHDATA2
    else:
        return b'\x4e' + (n).to_bytes(4, byteorder='little')    # OP_PUSHDATA4

def var_int(i):

    if i < 0xfd:
        return (i).to_bytes(1, byteorder='little')
    elif i <= 0xffff:
        return b'\xfd' + (i).to_bytes(2, byteorder='little')
    elif i <= 0xffffffff:
        return b'\xfe' + (i).to_bytes(4, byteorder='little')
    else:
        return b'\xff' + (i).to_bytes(8, byteorder='little')

def pack_uint64(i):
    upper = int(i / 4294967296)
    lower = i - upper * 4294967296

    return struct.pack('<L', lower) + struct.pack('<L', upper)

def monosig_script(address):
    '''returns a mono-signature output script'''

    hash160 = get_hash160(address)
    n = len(hash160)
    script = OP_DUP + OP_HASH160 + op_push(n) + hash160 + OP_EQUALVERIFY + OP_CHECKSIG
    return script

def op_return_script(data):
    '''returns a single OP_RETURN output script'''

    if not isinstance(data, bytes):
        data = data.encode()

    data = hexlify(data)
    script = hexlify(OP_RETURN + op_push(len(data)//2)) + data
    return unhexlify(script)

def make_raw_transaction(network, inputs, outputs, sequence_number=b'\xff\xff\xff\xff', lock_time=b'\x00\x00\x00\x00'):
    ''' inputs expected as [{'txid':txhash,'vout':index,'scriptSig':txinScript},..]
        ouputs expected as [{'redeem':peertoshis,'outputScript': outputScript},...]
    '''
    raw_tx = b'\x01\x00\x00\x00' # 4 byte version number
    network_vars = query(network)

    if network_vars.tx_timestamp:
        raw_tx += struct.pack('<L', int(time())) # 4 byte timestamp (Peercoin specific)

    raw_tx += var_int(len(inputs)) # varint for number of inputs

    for utxo in inputs:
        raw_tx += utxo['txid'][::-1] # previous transaction hash (reversed)
        raw_tx += struct.pack('<L', utxo['vout']) # previous txout index
        raw_tx += var_int(len(utxo['scriptSig'])) # scriptSig length
        raw_tx += utxo['scriptSig'] # scriptSig
        raw_tx += sequence_number # sequence number (irrelevant unless nLockTime > 0)

    raw_tx += var_int(len(outputs)) # varint for number of outputs

    for output in outputs:
        raw_tx += pack_uint64(int(round(output['redeem'] * network_vars.denomination ))) # value in peertoshi's or satoshi's
        raw_tx += var_int(len(output['outputScript']))
        raw_tx += output['outputScript']

    raw_tx += lock_time # nLockTime

    return raw_tx

def unpack_txn_buffer(buffer, network="ppc"):

    txn = {
        'vin': [],
        'vout': [],
        }

    txn['version'] = buffer.shift_unpack(4, '<L') # small-endian 32-bits
    if query(network).tx_timestamp: # peercoin: add 4 byte timestamp
        txn['timestamp'] = buffer.shift_unpack(4, '<L') # small-endian 32-bits

    inputs = buffer.shift_varint()
    if inputs > 100000: # sanity check
        return None

    for _ in range(inputs):
        _input = {}

        _input['txid'] = hexlify(buffer.shift(32)[::-1]).decode()
        _input['vout'] = buffer.shift_unpack(4, '<L')
        length = buffer.shift_varint()
        _input['scriptSig'] = hexlify(buffer.shift(length)).decode()
        _input['sequence'] = buffer.shift_unpack(4, '<L')

        txn['vin'].append(_input)

    outputs = buffer.shift_varint()
    if outputs > 100000: # sanity check
        return None

    for _ in range(outputs):
        output = {}

        output['value'] = float(buffer.shift_uint64()) / query(network).denomination
        length = buffer.shift_varint()
        output['scriptPubKey'] = hexlify(buffer.shift(length)).decode()

        txn['vout'].append(output)

    txn['locktime'] = buffer.shift_unpack(4, '<L')

    return txn

def unpack_raw_transaction(rawtx):
    '''unpacks raw transactions, returns dictionary'''

    if not isinstance(rawtx, bytes):
        raise ValueError('Binary input required')

    return unpack_txn_buffer(Tx_buffer(rawtx))




'''miscellaneous utilities.'''

import binascii
from pypeerassets.provider import RpcNode, Mintr
from pypeerassets.constants import *
from pypeerassets.main import Deck
from pypeerassets import paproto

def localnode_testnet_or_mainnet(node):
    '''check if local node is configured to testnet or mainnet'''

    if node.getinfo()["testnet"] is True:
        return "testnet"
    else:
        return "mainnet"

def load_p2th_privkeys_into_node(provider, prod=True):
    '''load production p2th privkey into local node'''

    assert isinstance(provider, RpcNode), {"error": "You can load privkeys only into local node."}
    error = {"error": "Loading P2TH privkey failed."}

    if provider.is_testnet:

        if prod is True:
            provider.importprivkey(testnet_PAPROD, "PAPROD")
            assert testnet_PAPROD_addr in provider.getaddressesbyaccount("PAPROD"), error

        else:
            provider.importprivkey(testnet_PATEST, "PATEST")
            assert testnet_PATEST_addr in provider.getaddressesbyaccount("PATEST"), error

    else:
        if prod is True:
            provider.importprivkey(mainnet_PAPROD, "PAPROD")
            assert mainnet_PAPROD_addr in provider.getaddressesbyaccount("PAPROD"), error

        else:
            provider.importprivkey(mainnet_PAPROD, "PAPROD")
            assert mainnet_PAPROD_addr in provider.getaddressesbyaccount("PAPROD"), error

def find_tx_sender(provider, txid):

    vin = provider.getrawtransaction(txid, 1)["vin"][0]
    txid = vin["txid"]
    index = vin["vout"]
    return provider.getrawtransaction(txid, 1)["vout"][index]["scriptPubKey"]["addresses"][0]

def find_deck_spawns(provider, prod=True):
    '''find deck spawn transactions via provider,
    it requires that Deck spawn P2TH were imported in local node or
    that remote API knows about P2TH address.'''

    if isinstance(provider, RpcNode):

        if prod:
            decks = [i["txid"] for i in provider.listtransactions("PAPROD")]
        else:
            decks = [i["txid"] for i in provider.listtransactions("PATEST")]

        return decks

    if isinstance(provider, Mintr):

        if prod:
            decks = [i["txid"] for i in provider.listtransactions(mainnet_PAPROD_addr)]
        else:
            raise NotImplementedError

def read_tx_opreturn(node, txid):
    '''Decode OP_RETURN message from <txid>'''

    vout = node.getrawtransaction(txid, 1)['vout'][1] # protocol requires that OP_RETURN is vout[1]

    asm = vout['scriptPubKey']['asm']
    n = asm.find('OP_RETURN')
    if n == -1:
        return False #{'error': 'OP_RETURN not found'}
    else:
        # add 10 because 'OP_RETURN ' is 10 characters
        n += 10
        data = asm[n:]
        n = data.find(' ')
        #make sure that we don't include trailing opcodes
        if n == -1:
            return binascii.unhexlify(data)
        else:
            return binascii.unhexlify(data[:n])

def validate_deckspawn_metainfo(deck):
    '''validate deck_spawn'''

    assert deck.version > 0, {"error": "Deck metainfo incomplete, version can't be 0."}
    assert deck.name is not "", {"error": "Deck metainfo incomplete, Deck must have a name."}
    assert deck.issue_mode in (0, 1, 2, 4), {"error": "Deck metainfo incomplete, unknown issue mode."}

def parse_deckspawn_metainfo(protobuf):
    '''decode deck_spawn tx op_return protobuf message and validate it.'''

    deck = paproto.DeckSpawn()
    deck.ParseFromString(protobuf)

    validate_deckspawn_metainfo(deck)

    return {
        "version": deck.version,
        "name": deck.name,
        "issue_mode": deck.MODE.Name(deck.issue_mode),
        "number_of_decimals": deck.number_of_decimals,
        "asset_specific_data": deck.asset_specific_data
    }

def validate_deckspawn_p2th(provider, deck_id, testnet=False, prod=True):
    '''validate if deck spawn pays to p2th in vout[0] and if it correct P2TH address'''

    raw = provider.getrawtransaction(deck_id, 1)
    vout = raw["vout"][0]["scriptPubKey"].get("addresses")[0]
    error = {"error": "This deck is not properly tagged."}

    if testnet:

        if not prod: # if test P2TH
            assert vout == testnet_PATEST_addr, error
            return True
        else:
            assert vout == testnet_PAPROD_addr, error
            return True

    else:

        if not prod: # if test P2TH
            assert vout == mainnet_PATEST_addr, error
            return True
        else:
            assert vout == mainnet_PAPROD_addr, error
            return True

def load_deck_p2th_into_local_node(provider, deck):
    '''
    load deck p2th into local node,
    this allows building of proof-of-timeline for this deck
    '''

    assert isinstance(deck, Deck), {"error": "Argument must be instance of PeerAssets Deck."}
    assert isinstance(provider, RpcNode), {"error": "You can load privkeys only into local node."}
    error = {"error": "Deck P2TH import went wrong."}

    provider.importprivkey(deck.p2th_wif, deck.name)
    assert deck.p2th_address in provider.getaddressesbyaccount(deck.name), error

def validate_card_transfer_p2th(provider, txid, deck):
    '''validate if card_transfer transaction pays to deck p2th in vout[0]'''

    raw = provider.getrawtransaction(txid, 1)
    error = {"error": "Card transfer is not properly tagged."}

    assert raw["vout"][0]["scriptPubKey"].get("addresses")[0] == deck.p2th_address, error

def parse_card_transfer_metainfo(protobuf):
    '''decode card_spawn tx op_return protobuf message and validate it.'''

    card = paproto.CardTransfer()
    card.ParseFromString(protobuf)

    assert card.version > 0, {"error": "Card metainfo incomplete, version can't be 0."}

    return {
        "version": card.version,
        "number_of_decimals": card.number_of_decimals,
        "amount": card.amounts,
        "asset_specific_data": card.asset_specific_data
    }





'''Kutil is for all things cryptography.'''

from hashlib import sha256, new
from binascii import hexlify, unhexlify
from base64 import b64encode, b64decode
from pypeerassets.base58 import b58encode, b58decode
from secp256k1 import PrivateKey
from pypeerassets import networks
from pypeerassets.transactions import var_int

class Kutil:
    '''pubkey/privkey operations'''

    def __init__(self, wif=None, privkey=None, network=None, seed=None):
        '''wif=<WIF> import private key from your wallet in WIF format
           privkey=<privkey> import private key in binary format
           network=<network> specify network [ppc, tppc, btc]
           seed=<seed> specify seed (string) to make the privkey from'''

        if wif is not None:
            key = self.wif_to_privkey(wif)
            self.keypair = PrivateKey(key['privkey'])
            network = key['net_prefix']

        if seed is not None:
            self.keypair = PrivateKey(privkey=self.seed_to_privkey(seed))

        if privkey is not None:
            self.keypair = PrivateKey(unhexlify(privkey))

        if privkey == wif == seed == None:
            self.keypair = PrivateKey() # new keypair

        if network is None:
            self.load_network_parameters('ppc')
        else:
            self.load_network_parameters(network)

        self._privkey = self.keypair.private_key
        self._pubkey = self.keypair.pubkey.serialize(compressed=False)
        self._pubkey_compressed = self.keypair.pubkey.serialize(compressed=True)

    def load_network_parameters(self, query):
        '''loads network parameters and sets class variables'''

        for field, var in zip(networks.query(query)._fields, networks.query(query)):
            setattr(self, field, var)

    @property
    def privkey(self):
        '''retrun privkey in hex format'''

        return hexlify(self._privkey).decode("utf-8")

    @property
    def pubkey(self):
        '''return compressed pubkey in hex format'''

        return hexlify(self._pubkey_compressed).decode("utf-8")

    @property
    def address(self):
        '''generate an address from pubkey'''

        key = self._pubkey_compressed # compressed pubkey as default

        keyhash = unhexlify(self.pubkeyhash + hexlify(
            new('ripemd160', sha256(key).digest()).
            digest())
                           )

        checksum = sha256(
            sha256(keyhash).digest()
            ).hexdigest()[0:8]

        return b58encode(keyhash + unhexlify(checksum))

    @staticmethod
    def check_wif(wif):
        '''check if WIF is properly formated.'''

        b58_wif = b58decode(wif)
        check = b58_wif[-4:]
        checksum = sha256(sha256(b58_wif[:-4]).digest()).digest()[0:4]
        return checksum == check

    @property
    def wif(self):
        '''convert raw private key to WIF'''

        extkey = self.wif_prefix + hexlify(self._privkey) + b'01' # compressed by default
        extcheck = unhexlify(extkey) + sha256(sha256(unhexlify(extkey
                                                              )).digest()).digest()[0:4]
        wif = b58encode(extcheck)

        assert self.check_wif(wif)
        return wif

    def wif_to_privkey(self, wif):
        '''import WIF'''

        assert self.check_wif(wif)
        b58_wif = b58decode(wif)

        if len(wif) == 51:
            return {'privkey': b58_wif[1:-4], 'net_prefix': hexlify(b58_wif[0:1])}
        if len(wif) == 52:
            return {'privkey': b58_wif[1:-5], 'net_prefix': hexlify(b58_wif[0:1])}
        else:
            return 'Invalid WIF length'

    @staticmethod
    def seed_to_privkey(seed):
        '''use mnemonic seed hash as privkey'''

        seed_hash = sha256(seed.encode("utf-8")).digest()
        return seed_hash

    def sign_message(self, message):
        '''sign >message< with the privkey and b64encode'''

        message = message.encode()
        prefix2 = var_int(len(message))
        buffer = b''.join([self.msgPrefix, prefix2, message])
        msg_hash = sha256(sha256(buffer).digest()).digest()
        sig = self.keypair.ecdsa_sign(msg_hash, raw=True)
        ## calculate the size of the key ?
        signature = self.keypair.ecdsa_serialize_compact(sig)

        return b64encode(var_int(len(signature)) + signature).decode()

    def ecdsa_sign(self, message):
        '''sign >message< with the privkey'''

        message = message.encode()
        msg_hash = sha256(sha256(message).digest()).digest()
        sig = self.keypair.ecdsa_sign(msg_hash, raw=True)

        return self.keypair.ecdsa_serialize_compact(sig)
    
    def verify_message(self, message, signature):
        '''verify >message< against >signature<'''

        sig = self.keypair.ecdsa_deserialize_compact(b64decode(message.encode("utf-8")))
        return self.keypair.pubkey.ecdsa_verify(sig, signature)




'''contains main protocol logic like assembly of proof-of-timeline and parsing deck info'''

import warnings
from binascii import hexlify, unhexlify
from pypeerassets import paproto, Kutil
from pypeerassets.pautils import *
from pypeerassets import constants, transactions
from .networks import query, networks

def find_all_valid_decks(provider, prod=True):
    '''
    scan the blockchain for PeerAssets decks, returns list of deck objects.
    please pass <node> - the provider
    <testnet> True/False
    <test> True/False - test or production P2TH
    '''

    decks = []
    deck_spawns = find_deck_spawns(provider) # find all deck_spawns on PAProd P2TH

    for i in deck_spawns:
        try:
            validate_deckspawn_p2th(provider, i, prod=prod, testnet=provider.is_testnet)
            if parse_deckspawn_metainfo(read_tx_opreturn(provider, i)):
                d = parse_deckspawn_metainfo(read_tx_opreturn(provider, i))
                d["asset_id"] = i
                try:
                    d["time"] = provider.getrawtransaction(i, 1)["blocktime"]
                except KeyError:
                    d["time"] = 0
                d["issuer"] = find_tx_sender(provider, i)
                d["network"] = provider.network
                decks.append(Deck(**d))

        except AssertionError:
            pass

    return decks

def find_deck(provider, key, prod=True):
    '''find specific deck by key, with key being:
    <id>, <name>, <issuer>, <issue_mode>, <number_of_decimals>
    '''

    decks = find_all_valid_decks(provider, prod=prod)
    return [d for d in decks if key in d.__dict__.values()]

class Deck:

    def __init__(self, name, number_of_decimals, issue_mode, network, version=1,
                 asset_specific_data="", issuer="", time=None, asset_id=None):
        '''
        initialize deck object, load from dictionary Deck(**dict)
        or initilize with kwargs Deck(1, "deck", 3, 2)
        '''

        self.version = version # protocol version
        self.name = name # deck name
        self.issue_mode = issue_mode # deck issue mode
        assert isinstance(number_of_decimals, int), {"error": "number_of_decimals must be an integer"}
        self.number_of_decimals = number_of_decimals
        self.asset_specific_data = asset_specific_data # optional metadata for the deck
        self.asset_id = asset_id
        self.issuer = issuer
        self.issue_time = time
        self.network = network
        if self.network.startswith("t"):
            self.testnet = True
        else:
            self.testnet = False

    @property
    def p2th_address(self):
        '''P2TH address of this deck'''

        return Kutil(network=self.network, privkey=self.asset_id).address

    @property
    def p2th_wif(self):
        '''P2TH privkey in WIF format'''

        return Kutil(network=self.network, privkey=self.asset_id).wif

    @property
    def metainfo_to_protobuf(self):
        '''encode deck into protobuf'''

        deck = paproto.DeckSpawn()
        deck.version = self.version
        deck.name = self.name
        deck.number_of_decimals = self.number_of_decimals
        deck.issue_mode = deck.MODE.Value(self.issue_mode)
        if not isinstance(self.asset_specific_data, bytes):
            deck.asset_specific_data = self.asset_specific_data.encode()
        else:
            deck.asset_specific_data = self.asset_specific_data

        proto = deck.SerializeToString()

        if len(proto) > 80:
            warnings.warn('\nMetainfo size exceeds maximum of 80bytes that fit into OP_RETURN.')

        return proto

    @property
    def metainfo_to_dict(self):
        '''encode deck into dictionary'''

        return {
            "version": self.version,
            "name": self.name,
            "number_of_decimals": self.number_of_decimals,
            "issue_mode": self.issue_mode
        }

def deck_spawn(deck, inputs, change_address, prod=True):
    '''spawn new deck, returns raw unsigned transaction'''

    network_params = query(deck.network)

    if deck.network.startswith("t"):
        p2th_fee = constants.testnet_p2th_fee
        if prod:
            p2th_address = constants.testnet_PAPROD_addr
        else:
            p2th_address = constants.testnet_PATEST_addr

    else:
        p2th_fee = constants.mainnet_p2th_fee
        if prod:
            p2th_address = constants.mainnet_PAPROD_addr
        else:
            p2th_address = constants.mainnet_PATEST_addr

    tx_fee = network_params.min_tx_fee # settle for min tx fee for now

    for utxo in inputs['utxos']:
        utxo['txid'] = unhexlify(utxo['txid'])
        utxo['scriptSig'] = unhexlify(utxo['scriptSig'])

    outputs = [
        {"redeem": p2th_fee, "outputScript": transactions.monosig_script(p2th_address)},
        {"redeem": 0, "outputScript": transactions.op_return_script(deck.metainfo_to_protobuf)},
        {"redeem": float(inputs['total']) - float(tx_fee) - float(p2th_fee), "outputScript": transactions.monosig_script(change_address)
        }]

    return transactions.make_raw_transaction(deck.network, inputs['utxos'], outputs)

def deck_transfer(deck, network, inputs, change_address, prod=True):
    '''
    The deck transfer transaction is a special case of the deck spawn transaction.
    Instead of registering a new asset, the deck transfer transaction transfers ownership from vin[1] to vin[0],
    meaning that both parties are required to sign the transfer transaction for it to be accepted in the blockchain.
    '''
    raise NotImplementedError

def find_all_card_transfers(provider, deck):
    '''find all <deck> card transfers'''

    cards = []
    card_transfers = provider.listtransactions(deck.name)

    nderror = {"error": "Number of decimals does not match."}

    for ct in card_transfers:
        try:
            validate_card_transfer_p2th(provider, ct["txid"], deck)

            if parse_card_transfer_metainfo(read_tx_opreturn(provider, ct["txid"])):

                raw_card = parse_card_transfer_metainfo(read_tx_opreturn(provider, ct["txid"]))
                _card = {}
                _card["version"] = raw_card["version"]
                _card["number_of_decimals"] = raw_card["number_of_decimals"]
                ## check if card number of decimals matches the deck atribute
                assert _card["number_of_decimals"] == deck.number_of_decimals, nderror

                _card["deck"] = deck
                _card["txid"] = ct["txid"]
                try:
                    _card["blockhash"] = ct["blockhash"]
                except KeyError:
                    _card["blockhash"] = 0
                _card["timestamp"] = ct["time"]
                _card["sender"] = find_tx_sender(provider, ct["txid"])
                _card["asset_specific_data"] = raw_card["asset_specific_data"]

                vouts = provider.getrawtransaction(ct["txid"], 1)["vout"]

                if len(raw_card["amount"]) > 1: ## if card states multiple outputs:
                    for am, v in zip(raw_card["amount"], vouts[2:]):
                        c = _card.copy()
                        c["amount"] = [am]
                        c["receivers"] = v["scriptPubKey"]["addresses"]
                        cards.append(CardTransfer(**c))
                else:
                    _card["receivers"] = vouts[2]["scriptPubKey"]["addresses"]
                    _card["amount"] = raw_card["amount"]
                    cards.append(CardTransfer(**_card))

        except AssertionError:
            pass

    return cards


class CardTransfer:

    def __init__(self, deck, receivers=[], amount=[], version=1, txid=None, sender=None, blockhash=None,
                 timestamp=None, asset_specific_data="", number_of_decimals=None):
        '''CardTransfer object, used when parsing card_transfers from the blockchain
        or when sending out new card_transfer.
        It can be initialized by passing the **kwargs and it will do the parsing,
        or it can be initialized with passed arguments.

        * deck - instance of Deck object
        * receivers - list of receivers
        * amounts - list of amounts to be sent, matching receivers
        * version - protocol version, default 1
        * txid - transaction ID of CardTransfer
        * sender - transaction sender
        * blockhash - block ID where the tx was first included
        * timestamp - unix timestamp of the block where it was first included
        * asset_specific_data - extra metadata
        * number_of_decimals - number of decimals for amount, inherited from Deck object'''

        self.version = version
        self.deck_id = deck.asset_id
        self.txid = txid
        self.sender = sender
        self.receivers = receivers
        self.amount = amount
        assert len(self.receivers) < 20, {"error": "Too many receivers."}
        assert len(self.amount) == len(self.receivers), {"error": "Amounts must match receivers."}
        if blockhash:
            self.blockhash = blockhash
        else:
            self.blockhash = 0
        self.timestamp = timestamp
        self.asset_specific_data = asset_specific_data
        self.p2th_address = deck.p2th_address
        if not number_of_decimals:
            self.number_of_decimals = deck.number_of_decimals

        assert str(self.amount)[::-1].find('.') <= deck.number_of_decimals, {"error": "Too many decimals."}

        if self.sender == deck.issuer:
            self.type = "CardIssue"
        elif self.receivers[0] == deck.issuer:
            self.type = "CardBurn"
        else:
            self.type = "CardTransfer"

    @property
    def metainfo_to_protobuf(self):
        '''encode card_transfer info to protobuf'''

        card = paproto.CardTransfer()
        card.version = self.version
        card.number_of_decimals = self.number_of_decimals
        if not isinstance(self.asset_specific_data, bytes):
            card.asset_specific_data = self.asset_specific_data.encode()
        else:
            card.asset_specific_data = self.asset_specific_data
        for i in self.amount:
            card.amounts.append(i)

        proto = card.SerializeToString()

        if len(proto) > 80:
            warnings.warn('\nMetainfo size exceeds maximum of 80bytes that fit into OP_RETURN.')

        return proto

def card_issue(deck, card_transfer, inputs, change_address, testnet=True, prod=True):
    '''issue cards for this deck
        Arguments:
        * deck - Deck object
        * card_transfer - CardTransfer object
        * inputs - utxo [has to be owned by deck issuer]
        * network - ppc/tppc
        * testnet - True/False
        * prod - production P2TH tag [True/False]
    '''

    issuer_error = {"error": "You must provide UTXO owned by the issuer of this deck."}

    network_params = query(deck.network)

    for utxo in inputs["utxos"]:
        assert utxo["address"] == deck.issuer, issuer_error

    if testnet:
        p2th_fee = constants.testnet_p2th_fee
    else:
        p2th_fee = constants.mainnet_p2th_fee

    tx_fee = network_params.min_tx_fee # settle for min tx fee for now

    for utxo in inputs['utxos']:
        utxo['txid'] = unhexlify(utxo['txid'])
        utxo['scriptSig'] = unhexlify(utxo['scriptSig'])

    outputs = [
        {"redeem": p2th_fee, "outputScript": transactions.monosig_script(deck.p2th_address)},
        {"redeem": 0, "outputScript": transactions.op_return_script(card_transfer.metainfo_to_protobuf)}
    ]

    for addr in card_transfer.receivers:
        outputs.append({"redeem": 0, "outputScript": transactions.monosig_script(addr)
                       })

    outputs.append(
        {"redeem": float(inputs['total']) - float(tx_fee) - float(p2th_fee),
         "outputScript": transactions.monosig_script(change_address)
        })

    return transactions.make_raw_transaction(deck.network, inputs['utxos'], outputs)

def card_burn(deck, card_transfer, inputs, change_address, testnet=True, prod=True):
    '''card burn transaction, cards are burned by sending the cards back to deck issuer'''

    assert deck.issuer in card_transfer.receivers, {"error": "One of the recipients must be deck issuer."}

    network_params = query(deck.network)

    if testnet:
        p2th_fee = constants.testnet_p2th_fee
    else:
        p2th_fee = constants.mainnet_p2th_fee

    tx_fee = network_params.min_tx_fee # settle for min tx fee for now

    for utxo in inputs['utxos']:
        utxo['txid'] = unhexlify(utxo['txid'])
        utxo['scriptSig'] = unhexlify(utxo['scriptSig'])

    outputs = [
        {"redeem": p2th_fee, "outputScript": transactions.monosig_script(deck.p2th_address)},
        {"redeem": 0, "outputScript": transactions.op_return_script(card_transfer.metainfo_to_protobuf)}
    ]

    for addr in card_transfer.receivers:
        outputs.append({"redeem": 0, "outputScript": transactions.monosig_script(addr)
                       })

    outputs.append(
        {"redeem": float(inputs['total']) - float(tx_fee) - float(p2th_fee),
         "outputScript": transactions.monosig_script(change_address)
        })

    return transactions.make_raw_transaction(deck.network, inputs['utxos'], outputs)

def card_transfer(deck, card_transfer, inputs, change_address, testnet=True, prod=True):
    '''standard peer-to-peer card transfer.'''

    network_params = query(deck.network)

    if testnet:
        p2th_fee = constants.testnet_p2th_fee
    else:
        p2th_fee = constants.mainnet_p2th_fee

    tx_fee = network_params.min_tx_fee # settle for min tx fee for now

    for utxo in inputs['utxos']:
        utxo['txid'] = unhexlify(utxo['txid'])
        utxo['scriptSig'] = unhexlify(utxo['scriptSig'])

    outputs = [
        {"redeem": p2th_fee, "outputScript": transactions.monosig_script(deck.p2th_address)},
        {"redeem": 0, "outputScript": transactions.op_return_script(card_transfer.metainfo_to_protobuf)}
    ]

    for addr in card_transfer.receivers:
        outputs.append({"redeem": 0, "outputScript": transactions.monosig_script(addr)
                       })

    outputs.append(
        {"redeem": float(inputs['total']) - float(tx_fee) - float(p2th_fee),
         "outputScript": transactions.monosig_script(change_address)
        })

    return transactions.make_raw_transaction(deck.network, inputs['utxos'], outputs)




from setuptools import setup

setup(name='pypeerassets',
      version='0.1',
      description='Python implementation of the PeerAssets protocol.',
      keywords=["blockchain", "digital asset", "protocol"],
      url='https://github.com/peerassets/pypeerassets',
      author='PeerAssets',
      author_email='peerchemist@protonmail.ch',
      license='GPL',
      packages=['pypeerassets', 'pypeerassets.provider'],
      install_requires=['requests', 'secp256k1', 'protobuf3'],
      zip_safe=True)
